#define BACKGROUND_C

#include "gba.h"
#include "GBA_Background.h"
#include "sincosrad.h"
#include "mem.h"

GBA_Background::GBA_Background(uint8 nBg)	// we need to know which we are
{
	m_nBg = nBg;
	m_pBGCNT = (&REG_BG0CNT)+nBg;
	m_pPOSDATA = (&REG_BG0HOFS)+(nBg*2);
	if (nBg == 2 && !(REG_DISPCNT & MODE0))
		m_pROTDATA = (&REG_BG2PA);
	else if (nBg == 3 && REG_DISPCNT & MODE2)
		m_pROTDATA = (&REG_BG3PA);
	else
		m_pROTDATA = 0;
}
GBA_Background::~GBA_Background()
{	// just turn it off (to be safe)
	REG_DISPCNT &= ~(BG0_ENABLE<<m_nBg);
}

// General methods
bool GBA_Background::IsEnabled()
{
	return ITRUE(REG_DISPCNT & (BG0_ENABLE<<m_nBg));
}
void GBA_Background::Enable()
{
	REG_DISPCNT |= (BG0_ENABLE<<m_nBg);
}
void GBA_Background::Disable()
{
	REG_DISPCNT &= ~(BG0_ENABLE<<m_nBg);
}

// Specific methods
bool GBA_Background::IsMosaic()
{
	return ITRUE(*m_pBGCNT & BG_MOSAIC_ENABLE);
}
void GBA_Background::EnableMosaic()
{
	*m_pBGCNT |= BG_MOSAIC_ENABLE;
}
void GBA_Background::DisableMosaic()
{
	*m_pBGCNT &= ~BG_MOSAIC_ENABLE;
}

// Set with low number (0-3)
void GBA_Background::SetSize(uint8 nSize)
{
	*m_pBGCNT &= ~(3<<14);	// clear the size
	*m_pBGCNT |= ((nSize&3)<<14); // reset it
}
uint8 GBA_Background::GetSize()
{
	return (REG_BG0CNT>>14) & 3;
}

// Tile and level map base get/set methods (0-3)
void GBA_Background::SetTileSetBase(uint8 nBase)
{
	*m_pBGCNT &= ~(3<<2); // clear old value
	*m_pBGCNT |= ((nBase&3)<<2); // Set new value

	m_Base &= ~3;
	m_Base |= nBase&3;
}
uint8 GBA_Background::GetTileSetBase()
{
	return (*m_pBGCNT>>2) & 3;
}
// level map (0-31)
void GBA_Background::SetLevelMapBase(uint8 nBase)
{
	*m_pBGCNT &= ~(31<<8); // clear old value
	*m_pBGCNT |= ((nBase&31)<<8); // Set new value

	m_Base &= 3;
	m_Base |= (nBase&31)<<2;
}
uint8 GBA_Background::GetLevelMapBase()
{
	return (*m_pBGCNT>>8) & 31;
}

// Color Depth Get/Set Methods
// Can be either defines or 0-1
void GBA_Background::SetColorDepth(uint8 nDepth)
{
	*m_pBGCNT &= ~(1<<7); // clear old value
	*m_pBGCNT |= ((nDepth&1)<<7); // Set new value
}
uint8 GBA_Background::GetColorDepth()
{
	return (REG_BG0CNT>>7) & 1;
}

// Priority Get/Sets - (0-3)
void GBA_Background::SetPriority(uint8 nPri)
{
	*m_pBGCNT &= ~3; // clear old value
	*m_pBGCNT |= nPri&3; // Set new value
}
uint8 GBA_Background::GetPriority()
{
	return *m_pBGCNT & 3;
}

// Interaction Functions
void GBA_Background::ScrollMap(int16 x, int16 y)
{
	m_X += x;
	m_Y += y;
}
void GBA_Background::SetOffset(int16 x, int16 y)
{
	m_X = x;
	m_Y = y;
}

// Rotate is comulative
void GBA_Background::Rotate(int16 angle)
{
	m_Rot += angle;
	RecalcRot();
}
// Set and gets for rotation
void GBA_Background::SetRotation(int16 angle)
{
	m_Rot = angle;
	RecalcRot();
}
int16 GBA_Background::GetRotation()
{
	return m_Rot;
}

// Sets and gets for the center of the map
void GBA_Background::SetCenter(int16 x, int16 y)
{
	m_CentreX = x;
	m_CentreY = y;
}
void GBA_Background::GetCenter(int16* x, int16* y)
{
	*x = m_CentreX;
	*y = m_CentreY;
}

// Zoom Get/Sets
void GBA_Background::SetZoom(int32 zoom)
{
	m_Zoom = zoom;
}
int32 GBA_Background::GetZoom()
{
	return m_Zoom;
}
void GBA_Background::CopyTileData(uint16* pSource, uint16 nCount)
{
	if (GetColorDepth())
		memcpy((uint16*)CharBaseBlock(GetTileSetBase()), pSource, nCount*2);
	else
		memcpy((uint16*)CharBaseBlock(GetTileSetBase()), pSource, nCount);
}
void GBA_Background::CopyMapData(uint16* pSource, uint16 nCount)
{
	memcpy((uint16*)ScreenBaseBlock(GetLevelMapBase()), pSource, nCount*2);
}
void GBA_Background::CopyRow(uint16* pSource, uint8 sx, uint8 sy, uint8 nLen)
{
	uint16* pMapData = (uint16*)ScreenBaseBlock(GetLevelMapBase());
	for (int cnt = 0; cnt < nLen; cnt++)
	{
		pMapData[sx + (sy*32)] = pSource[cnt];
		++sx;
		if (sx > 31)
			sx = 0;
	}
}
void GBA_Background::CopyColumn(uint16* pSource, uint8 sx, uint8 sy, uint8 nLen)
{
	uint16* pMapData = (uint16*)ScreenBaseBlock(GetLevelMapBase());
	for (int cnt = 0; cnt < nLen; cnt++)
	{
		pMapData[sx + (sy*32)] = pSource[cnt];
		++sy;
		if (sy > 31)
			sy = 0;
	}
}

void GBA_Background::Update()
{
	if (m_pROTDATA)	// rot background
	{
		*(m_pROTDATA) = PA;
		*(m_pROTDATA+1) = PB;
		*(m_pROTDATA+2) = PC;
		*(m_pROTDATA+3) = PD;

		*(u32*)(m_pROTDATA+4) = DX;
		*(u32*)(m_pROTDATA+6) = DY;
	}
	else //it is a text background
	{
		*(m_pPOSDATA) = m_X;
		*(m_pPOSDATA+1) = m_Y;
	}
}

void GBA_Background::RecalcRot()
{
	// clamp rotation so we don't reference data outside of cos/sin
	while (m_Rot > 360)
		m_Rot -= 360;
	while (m_Rot < 0)
		m_Rot += 360;

	int16 center_y = (int16)((int32)m_CentreY * m_Zoom)>>8;
	int16 center_x = (int16)((int32)m_CentreX * m_Zoom)>>8;
	
	DX = ((m_X<<8)-center_y*SIN[m_Rot]-center_x*COS[m_Rot]);
	DY = ((m_Y<<8)-center_y*COS[m_Rot]+center_x*SIN[m_Rot]);
	
	PA = ((int32)COS[m_Rot]*(int32)m_Zoom)>>8;  //cos&sin are LUTs that are .8 fixed numbers
	PB = ((int32)SIN[m_Rot]*(int32)m_Zoom)>>8;  //zoom is also fixed
	PC = (-(int32)SIN[m_Rot]*(int32)m_Zoom)>>8;
	PD = ((int32)COS[m_Rot]*(int32)m_Zoom)>>8;
}


